@@ -4,7 +4,6 @@ import abc
 import asyncio
 import dataclasses
 import ipaddress
-import threading
 from typing import (
     Any,
     Awaitable,
@@ -15,8 +14,10 @@ from typing import (
     Set,
     Tuple,
     TypeVar,
+    cast,
 )
 
+import aiobotocore.session
 import boto3
 
 from meadowrun.aws_integration.aws_core import (
@@ -30,6 +31,7 @@ from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
     _EC2_ALLOC_TAG,
     _EC2_ALLOC_TAG_VALUE,
     ignore_boto3_error_code,
+    ignore_boto3_error_code_async,
 )
 from meadowrun.aws_integration.quotas import (
     get_quota_for_instance_type,
@@ -249,12 +251,22 @@ class LaunchEC2InstanceSuccess(LaunchEC2InstanceResult):
     """
     Indicates that launch_ec2_instance succeeded. public_address_continuation can be
     awaited to get the public address of the launched instance.
+    public_address_continuation can be None which indicates that the launch of the
+    instance was cancelled. If the instance was mid-launch, then the instance was
+    terminated before it could fully launch.
     """
 
-    public_address_continuation: Awaitable[str]
+    public_address_continuation: Awaitable[Optional[str]]
     instance_id: str
 
 
+@dataclasses.dataclass(frozen=True)
+class LaunchEC2InstanceAborted(LaunchEC2InstanceResult):
+    """Indicates that launch_ec2_instance was aborted"""
+
+    pass
+
+
 class LaunchEC2InstanceTypeUnavailable(LaunchEC2InstanceResult):
     """
     Indicates that launch_ec2_instance failed because the specified instance type is not
@@ -320,17 +332,29 @@ class LaunchEC2InstanceQuota(LaunchEC2InstanceTypeUnavailable):
         )
 
 
+@dataclasses.dataclass(frozen=True)
+class LaunchEC2InstanceSettings:
+    """
+    This class contains the settings that come from AllocEC2Instance. See that class for
+    details on the semantics of each argument.
+    """
+
+    ami_id: str
+    subnet_id: Optional[str] = None
+    security_group_ids: Optional[Sequence[str]] = None
+    iam_role_instance_profile: Optional[str] = None
+
+
 async def launch_ec2_instance(
     region_name: str,
     instance_type: str,
     on_demand_or_spot: OnDemandOrSpotType,
-    ami_id: str,
-    security_group_ids: Optional[Sequence[str]] = None,
-    iam_role_name: Optional[str] = None,
+    launch_settings: LaunchEC2InstanceSettings,
     user_data: Optional[str] = None,
     key_name: Optional[str] = None,
     tags: Optional[Dict[str, str]] = None,
     volume_size_gb: int = 100,
+    abort: Optional[asyncio.Event] = None,
 ) -> LaunchEC2InstanceResult:
     """
     Launches the specified EC2 instance. See derived classes of LaunchEC2InstanceResult
@@ -350,10 +374,16 @@ async def launch_ec2_instance(
             }
         ]
     }
-    if security_group_ids:
-        optional_args["SecurityGroupIds"] = security_group_ids
-    if iam_role_name:
-        optional_args["IamInstanceProfile"] = {"Name": iam_role_name}
+    if launch_settings.subnet_id:
+        # this should not be set if we specify a custom network interface (which isn't
+        # supported yet)
+        optional_args["SubnetId"] = launch_settings.subnet_id
+    if launch_settings.security_group_ids:
+        optional_args["SecurityGroupIds"] = launch_settings.security_group_ids
+    if launch_settings.iam_role_instance_profile:
+        optional_args["IamInstanceProfile"] = {
+            "Name": launch_settings.iam_role_instance_profile
+        }
     if key_name:
         optional_args["KeyName"] = key_name
     if user_data:
@@ -373,101 +403,157 @@ async def launch_ec2_instance(
     else:
         raise ValueError(f"Unexpected value for on_demand_or_spot {on_demand_or_spot}")
 
-    ec2_resource = boto3.resource("ec2", region_name=region_name)
-    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.run_instances
-    success, instances, error_code = ignore_boto3_error_code(
-        lambda: ec2_resource.create_instances(
-            ImageId=ami_id,
-            MinCount=1,
-            MaxCount=1,
-            InstanceType=instance_type,
-            **optional_args,
-        ),
-        {
-            "InsufficientInstanceCapacity",
-            "MaxSpotInstanceCountExceeded",
-            "VcpuLimitExceeded",
-        },
-        True,
-    )
-    if not success:
-        if error_code == "InsufficientInstanceCapacity":
-            return LaunchEC2InstanceAwsCapacity(
-                f"AWS does not have enough capacity in region {region_name} to create a"
-                f" {on_demand_or_spot} instance of type {instance_type}",
-                instance_type,
-                on_demand_or_spot,
-            )
-        elif error_code in ("MaxSpotInstanceCountExceeded", "VcpuLimitExceeded"):
-            # I believe these error codes have the same meaning for spot/on-demand
-            # instances
-            return LaunchEC2InstanceQuota(
-                *get_quota_for_instance_type(
-                    instance_type, on_demand_or_spot, region_name
+    if abort is not None and abort.is_set():
+        return LaunchEC2InstanceAborted()
+
+    async with aiobotocore.session.get_session().create_client(
+        "ec2", region_name=region_name
+    ) as ec2_client:
+        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.run_instances
+        success, instances, error_code = await ignore_boto3_error_code_async(
+            ec2_client.run_instances(
+                ImageId=launch_settings.ami_id,
+                MinCount=1,
+                MaxCount=1,
+                InstanceType=instance_type,  # type: ignore
+                **optional_args,
+            ),
+            {
+                "InsufficientInstanceCapacity",
+                "MaxSpotInstanceCountExceeded",
+                "VcpuLimitExceeded",
+            },
+        )
+        if not success:
+            if error_code == "InsufficientInstanceCapacity":
+                return LaunchEC2InstanceAwsCapacity(
+                    f"AWS does not have enough capacity in region {region_name} to "
+                    f"create a {on_demand_or_spot} instance of type {instance_type}",
+                    instance_type,
+                    on_demand_or_spot,
                 )
-            )
-        else:
-            raise ValueError(f"Unexpected boto3 error code {error_code}")
+            elif error_code in ("MaxSpotInstanceCountExceeded", "VcpuLimitExceeded"):
+                # I believe these error codes have the same meaning for spot/on-demand
+                # instances
+                return LaunchEC2InstanceQuota(
+                    *get_quota_for_instance_type(
+                        instance_type, on_demand_or_spot, region_name
+                    )
+                )
+            else:
+                raise ValueError(f"Unexpected boto3 error code {error_code}")
 
-    assert instances is not None  # just for mypy
+    # types_aiobotocore seems to have the wrong types for instances
+    instances_untyped = cast(Any, instances)
+    if len(instances_untyped["Instances"]) != 1:
+        raise ValueError(
+            "run_instances succeeded but returned "
+            f"{len(instances_untyped['Instances'])} instances"
+        )
+    instance_id = instances_untyped["Instances"][0]["InstanceId"]
     return LaunchEC2InstanceSuccess(
-        _launch_instance_continuation(instances[0]), instances[0].id
+        _launch_instance_continuation(instance_id, region_name, abort), instance_id
     )
 
 
-async def _launch_instance_continuation(instance: Any) -> str:
+async def describe_single_instance(ec2_client: Any, instance_id: str) -> Dict:
+    response = await ec2_client.describe_instances(InstanceIds=[instance_id])
+    reservations = response["Reservations"]
+    if len(reservations) != 1:
+        raise ValueError(f"Unexpected number of reservations: {len(reservations)}")
+    instances = reservations[0]["Instances"]
+    if len(instances) != 1:
+        raise ValueError(f"Unexpected number of instances: {len(instances)}")
+    return instances[0]
+
+
+async def _launch_instance_continuation(
+    instance_id: str, region_name: str, abort: Optional[asyncio.Event]
+) -> Optional[str]:
     """
     instance should be a boto3 EC2 instance, waits for the instance to be running and
     then returns its public DNS name
     """
 
-    # boto3 doesn't have any async APIs, which means that in order to run more than one
-    # launch_ec2_instance at the same time, we need to have a thread that waits. We use
-    # an asyncio.Future here to make the API async, so from the user perspective, it
-    # feels like this function is async
-
-    # boto3 should be threadsafe:
-    # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/clients.html#multithreading-or-multiprocessing-with-clients
-    instance_running_future: asyncio.Future = asyncio.Future()
-    event_loop = asyncio.get_running_loop()
-
-    def wait_until_running() -> None:
+    # ideally we would reuse the ec2_client from the caller, but this is much simpler
+    i = 0
+    async with aiobotocore.session.get_session().create_client(
+        "ec2", region_name=region_name
+    ) as ec2_client:
         try:
-            instance.wait_until_running()
-            event_loop.call_soon_threadsafe(
-                lambda: instance_running_future.set_result(None)
-            )
-        except asyncio.CancelledError:
+            # this is a reimplementation of
+            # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Instance.wait_until_running
+            # main difference is that we check every 7 seconds rather than every 15
+            # seconds
+            while True:
+                await asyncio.sleep(7)
+                # ideally we would interrupt the above sleep if cancel becomes set
+                if abort is not None and abort.is_set():
+                    await ec2_client.terminate_instances(InstanceIds=[instance_id])
+                    return None
+
+                instance = await describe_single_instance(ec2_client, instance_id)
+                if instance["State"]["Name"] == "running":
+                    break
+                elif instance["State"]["Name"] == "pending":
+                    pass
+                else:
+                    print(
+                        f"Unexpected instance state while waiting for {instance_id} to "
+                        f"start up: {instance['State']}. This instance may never start "
+                        "up."
+                    )
+
+                i += 1
+                if i >= 60:
+                    raise TimeoutError(
+                        "Checked 60 times, waiting 7 seconds between each check and "
+                        f"instance {instance_id} is still not running"
+                    )
+
+            if instance["PublicDnsName"]:
+                dns_name = instance["PublicDnsName"]
+            else:
+                # try one more time after waiting 1 seconds
+                await asyncio.sleep(1.0)
+                instance = await describe_single_instance(ec2_client, instance_id)
+                if instance["PublicDnsName"]:
+                    dns_name = instance["PublicDnsName"]
+                elif instance["PrivateDnsName"]:
+                    print(
+                        "Warning, no public DNS name available for instance "
+                        f"{instance_id}. Will try to use the private DNS name, but this"
+                        " will only work if the current machine is in the same VPC as "
+                        "the new instance. If this message is unexpected, please try "
+                        "setting your subnet to automatically assign public IPv4 "
+                        "addresses"
+                    )
+                    dns_name = instance["PrivateDnsName"]
+                else:
+                    raise ValueError(
+                        f"Instance {instance_id} is running, but does not have a public"
+                        " DNS name or a private DNS name."
+                    )
+        except BaseException:
+            # This is a best effort at terminating the instance. If we don't manage to
+            # terminate, it will get cleaned up by the management lambda as the instance
+            # will exist but not be registered. There's code before and after this that
+            # we should also wrap in a similar try/except block, but this is by far the
+            # most likely place where there will be an issue.
+            await ec2_client.terminate_instances(InstanceIds=[instance_id])
             raise
-        except Exception as e:
-            exception = e
-            event_loop.call_soon_threadsafe(
-                lambda: instance_running_future.set_exception(exception)
-            )
-
-    threading.Thread(target=wait_until_running).start()
-    await instance_running_future
-
-    instance.load()
-    if not instance.public_dns_name:
-        # try one more time after waiting 1 seconds
-        await asyncio.sleep(1.0)
-        instance.load()
-        if not instance.public_dns_name:
-            raise ValueError("Waited until running, but still no IP address!")
-    return instance.public_dns_name
+    return dns_name
 
 
 async def launch_ec2_instances(
-    resources_required_per_job: ResourcesInternal,
+    instance_type_resources_required_per_job: ResourcesInternal,
     num_jobs: int,
-    ami_id: str,
+    launch_settings: LaunchEC2InstanceSettings,
     region_name: Optional[str] = None,
-    security_group_ids: Optional[Sequence[str]] = None,
-    iam_role_name: Optional[str] = None,
     user_data: Optional[str] = None,
     key_name: Optional[str] = None,
     tags: Optional[Dict[str, str]] = None,
+    abort: Optional[asyncio.Event] = None,
 ) -> Sequence[CloudInstance]:
     """
     Launches enough EC2 instances to run num_jobs jobs that each require the specified
@@ -501,7 +587,7 @@ async def launch_ec2_instances(
             )
 
         chosen_instance_types = choose_instance_types_for_job(
-            resources_required_per_job,
+            instance_type_resources_required_per_job,
             num_jobs_left_to_allocate,
             (
                 i
@@ -513,18 +599,20 @@ async def launch_ec2_instances(
         at_least_one_chosen_instance_type = False
         num_jobs_left_to_allocate = 0
         for instance_type in chosen_instance_types:
+            if abort is not None and abort.is_set():
+                break
+
             at_least_one_chosen_instance_type = True
             for _ in range(instance_type.num_instances):
                 launch_ec2_result = await launch_ec2_instance(
                     region_name,
                     instance_type.instance_type.name,
                     instance_type.instance_type.on_demand_or_spot,
-                    ami_id=ami_id,
-                    security_group_ids=security_group_ids,
-                    iam_role_name=iam_role_name,
+                    launch_settings,
                     user_data=user_data,
                     key_name=key_name,
                     tags=tags,
+                    abort=abort,
                 )
                 if isinstance(launch_ec2_result, LaunchEC2InstanceSuccess):
                     public_dns_name_tasks.append(
@@ -532,6 +620,8 @@ async def launch_ec2_instances(
                     )
                     instance_ids.append(launch_ec2_result.instance_id)
                     launched_instance_types_repeated.append(instance_type)
+                elif isinstance(launch_ec2_result, LaunchEC2InstanceAborted):
+                    break
                 elif isinstance(launch_ec2_result, LaunchEC2InstanceTypeUnavailable):
                     print(launch_ec2_result.get_message())
 
@@ -551,7 +641,7 @@ async def launch_ec2_instances(
         if not at_least_one_chosen_instance_type:
             raise ValueError(
                 "There were no instance types that could be selected for: "
-                f"{resources_required_per_job}"
+                f"{instance_type_resources_required_per_job}"
             )
 
     public_dns_names = await asyncio.gather(*public_dns_name_tasks)
@@ -561,6 +651,8 @@ async def launch_ec2_instances(
         for public_dns_name, instance_id, instance_type in zip(
             public_dns_names, instance_ids, launched_instance_types_repeated
         )
+        # None indicates that the instance launch was cancelled
+        if public_dns_name is not None
     ]
 
 
