@@ -1,3 +1,6 @@
+from __future__ import annotations
+
+import dataclasses
 import hashlib
 from io import BytesIO
 from typing import Any, Dict, List, Optional, Tuple
@@ -15,6 +18,7 @@ from meadowrun.aws_integration.aws_core import (
 from meadowrun.aws_integration.management_lambdas.ec2_alloc_stub import (
     ignore_boto3_error_code,
 )
+from meadowrun.run_job_core import S3CompatibleObjectStorage
 
 BUCKET_PREFIX = "meadowrun"
 
@@ -155,9 +159,6 @@ async def list_objects_async(
 ) -> List[str]:
     """Returns the keys in the meadowrun bucket."""
     bucket_name = _get_bucket_name(region_name)
-    # response = await s3_client.list_objects_v2(
-    #     Bucket=bucket_name, Prefix=prefix, StartAfter=start_after
-    # )
 
     paginator = s3_client.get_paginator("list_objects_v2")
     results = []
@@ -168,13 +169,6 @@ async def list_objects_async(
             results.append(c["Key"])
     return results
 
-    # if response["KeyCount"] == 0:
-    #     return tuple(), False
-    # return (
-    #     tuple(content["Key"] for content in response["Contents"]),
-    #     response["IsTruncated"],
-    # )
-
 
 def download(
     object_name: str, region_name: str, byte_range: Optional[Tuple[int, int]] = None
@@ -224,3 +218,22 @@ def delete_bucket(region_name: str) -> None:
             bucket.delete_objects(Delete=dict(Objects=key_chunk))
 
         bucket.delete()
+
+
+@dataclasses.dataclass
+class S3ObjectStorage(S3CompatibleObjectStorage):
+    region_name: Optional[str] = None
+
+    @classmethod
+    def get_url_scheme(cls) -> str:
+        return "s3"
+
+    async def _upload(self, file_path: str) -> Tuple[str, str]:
+        return await ensure_uploaded(file_path)
+
+    async def _download(
+        self, bucket_name: str, object_name: str, file_name: str
+    ) -> None:
+        return await download_file(
+            bucket_name, object_name, file_name, self.region_name
+        )
