@@ -1,13 +1,25 @@
 from __future__ import annotations
 
+import asyncio
 import dataclasses
 import datetime
+import functools
 import json
-from typing import TYPE_CHECKING, Any, Iterable, List, Optional, Sequence, Tuple, Type
-
-if TYPE_CHECKING:
-    from typing_extensions import Literal
-    from types import TracebackType
+import uuid
+from typing import (
+    Any,
+    AsyncIterable,
+    Callable,
+    Generic,
+    Iterable,
+    List,
+    Optional,
+    Sequence,
+    TYPE_CHECKING,
+    Tuple,
+    Type,
+    TypeVar,
+)
 
 import meadowrun.azure_integration.azure_vms
 from meadowrun.azure_integration.azure_meadowrun_core import (
@@ -15,6 +27,16 @@ from meadowrun.azure_integration.azure_meadowrun_core import (
     get_default_location,
 )
 from meadowrun.azure_integration.azure_ssh_keys import ensure_meadowrun_key_pair
+from meadowrun.azure_integration.blob_storage import AzureBlobStorage
+from meadowrun.azure_integration.grid_tasks_queue import (
+    Queue,
+    _add_tasks,
+    _add_worker_shutdown_message,
+    _create_queues_for_job,
+    _retry_task,
+    get_results_unordered,
+    worker_function,
+)
 from meadowrun.azure_integration.mgmt_functions.azure_constants import (
     ALLOCATED_TIME,
     LAST_UPDATE_TIME,
@@ -42,14 +64,28 @@ from meadowrun.instance_allocation import (
     InstanceRegistrar,
     _InstanceState,
     _TInstanceState,
-    allocate_jobs_to_instances,
+    allocate_single_job_to_instance,
 )
 from meadowrun.instance_selection import CloudInstance, ResourcesInternal
+from meadowrun.run_job_core import (
+    AllocVM,
+    CloudProviderType,
+    GridJobCloudInterface,
+    JobCompletion,
+    ObjectStorage,
+    SshHost,
+    WaitOption,
+)
 
 if TYPE_CHECKING:
-    from meadowrun.meadowrun_pb2 import Job
+    import asyncssh
+    from typing_extensions import Literal
+    from types import TracebackType
+    from meadowrun.meadowrun_pb2 import Job, ProcessState
 
-from meadowrun.run_job_core import JobCompletion, SshHost
+
+_T = TypeVar("_T")
+_U = TypeVar("_U")
 
 
 @dataclasses.dataclass
@@ -374,8 +410,17 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
         self,
         resources_required_per_task: ResourcesInternal,
         num_concurrent_tasks: int,
-        region_name: str,
+        alloc_cloud_instances: AllocVM,
+        abort: Optional[asyncio.Event],
     ) -> Sequence[CloudInstance]:
+        if not isinstance(alloc_cloud_instances, AllocAzureVM):
+            # TODO do this in the type checker somehow
+            raise ValueError(
+                "Programming error: AzureInstanceRegistrar can only be used with "
+                "AllocAzureVM"
+            )
+
+        region_name = alloc_cloud_instances._get_location()
         return await meadowrun.azure_integration.azure_vms.launch_vms(
             resources_required_per_task,
             num_concurrent_tasks,
@@ -395,45 +440,201 @@ class AzureInstanceRegistrar(InstanceRegistrar[AzureVMInstanceState]):
     ) -> None:
         if ports:
             raise NotImplementedError(
-                "Opening ports on Azure is not implemented, please create an issue at "
-                "https://github.com/meadowdata/meadowrun/issues"
+                "Opening ports on Azure is not implemented, please comment on "
+                "https://github.com/meadowdata/meadowrun/issues/126"
             )
 
 
+@dataclasses.dataclass
+class AllocAzureVM(AllocVM):
+    """
+    Specifies that the job should be run on a dynamically allocated Azure VM. Any
+    existing Meadowrun-managed VMs will be reused if available. If none are available,
+    Meadowrun will launch the cheapest VM type that meets the resource requirements for
+    a job.
+
+    `resources_required` must be provided with the AllocAzureVM Host.
+
+    Attributes:
+        location: Specifies the location for the Azure VM, e.g. "eastus". None will use
+            the default location.
+    """
+
+    location: Optional[str] = None
+
+    def get_cloud_provider(self) -> CloudProviderType:
+        return "AzureVM"
+
+    async def set_defaults(self) -> None:
+        self.location = get_default_location()
+
+    def _get_location(self) -> str:
+        # Small wrapper around region_name that throws if it is None. Should only be
+        # used internally.
+        if self.location is None:
+            raise ValueError(
+                "Programming error: location is None but it should have been set by "
+                "set_defaults earlier"
+            )
+        return self.location
+
+    def get_runtime_resources(self) -> ResourcesInternal:
+        return ResourcesInternal({}, {})
+
+    async def run_job(
+        self,
+        resources_required: Optional[ResourcesInternal],
+        job: Job,
+        wait_for_result: WaitOption,
+    ) -> JobCompletion[Any]:
+        if resources_required is None:
+            raise ValueError(
+                "Resources.logical_cpu and memory_gb must be specified for "
+                "AllocAzureVM"
+            )
+
+        return await run_job_azure_vm_instance_registrar(
+            job, resources_required, self, wait_for_result
+        )
+
+    def _create_grid_job_cloud_interface(self) -> GridJobCloudInterface:
+        return AzureVMGridJobInterface(self)
+
+    async def get_object_storage(self) -> ObjectStorage:
+        return AzureBlobStorage()
+
+
 async def run_job_azure_vm_instance_registrar(
     job: Job,
     resources_required: ResourcesInternal,
-    location: Optional[str],
-    wait_for_result: bool,
+    alloc_azure_vm: AllocAzureVM,
+    wait_for_result: WaitOption,
 ) -> JobCompletion[Any]:
-    if not location:
-        location = get_default_location()
+    location = alloc_azure_vm._get_location()
+
     pkey, public_key = await ensure_meadowrun_key_pair(location)
 
     async with AzureInstanceRegistrar(location, "create") as instance_registrar:
         if job.ports:
             raise NotImplementedError(
-                "Support for opening ports on Azure is not yet implemented, please "
-                "create an issue at https://github.com/meadowdata/meadowrun/issues"
+                "Opening ports on Azure is not implemented, please comment on "
+                "https://github.com/meadowdata/meadowrun/issues/126"
             )
-        hosts = await allocate_jobs_to_instances(
+        host, job_id = await allocate_single_job_to_instance(
             instance_registrar,
             resources_required,
-            1,
-            location,
+            alloc_azure_vm,
             job.ports,
         )
 
-    if len(hosts) != 1:
-        raise ValueError(f"Asked for one host, but got back {len(hosts)}")
-    host, job_ids = list(hosts.items())[0]
-    if len(job_ids) != 1:
-        raise ValueError(f"Asked for one job allocation but got {len(job_ids)}")
-
     # Kind of weird that we're changing the job_id here, but okay as long as job_id
     # remains mostly an internal concept
-    job.job_id = job_ids[0]
+    job.job_id = job_id
 
     return await SshHost(host, "meadowrunuser", pkey, ("AzureVM", location)).run_job(
         resources_required, job, wait_for_result
     )
+
+
+class AzureVMGridJobInterface(GridJobCloudInterface, Generic[_T, _U]):
+    """
+    This is a relatively thin wrapper around some functionality in grid_tasks_queue.
+    This class should be in grid_tasks_queue, but it's here because of circular import
+    issues.
+    """
+
+    def __init__(self, alloc_cloud_instance: AllocAzureVM):
+        self._cloud_provider = alloc_cloud_instance.get_cloud_provider()
+        self._location = alloc_cloud_instance._get_location()
+
+        self._ssh_private_key: Optional[
+            asyncio.Task[Tuple[asyncssh.SSHKey, str]]
+        ] = None
+
+        self._request_result_queues: Optional[asyncio.Task[Tuple[Queue, Queue]]] = None
+
+        self._tasks: Optional[Sequence[_T]] = None
+
+        # this id is just used for creating the job's queues. It has no relationship to
+        # any Job.job_ids
+        self._job_id = str(uuid.uuid4())
+
+    def create_instance_registrar(self) -> InstanceRegistrar:
+        return AzureInstanceRegistrar(self._location, "create")
+
+    async def setup_and_add_tasks(self, tasks: Sequence[_T]) -> None:
+        print(f"The current run_map's id is {self._job_id}")
+        self._ssh_private_key = asyncio.create_task(
+            ensure_meadowrun_key_pair(self._location)
+        )
+        self._request_result_queues = asyncio.create_task(
+            _create_queues_for_job(self._job_id, self._location)
+        )
+        self._tasks = tasks
+        await _add_tasks((await self._request_result_queues)[0], tasks)
+
+    async def ssh_host_from_address(self, address: str) -> SshHost:
+        if self._ssh_private_key is None:
+            raise ValueError(
+                "Must call setup_and_add_tasks before calling ssh_host_from_address"
+            )
+
+        return SshHost(
+            address,
+            "meadowrunuser",
+            (await self._ssh_private_key)[0],
+            (self._cloud_provider, self._location),
+        )
+
+    async def shutdown_workers(self, num_workers: int) -> None:
+        if self._request_result_queues is None:
+            raise ValueError(
+                "Must call setup_and_add_tasks before calling shutdown_workers"
+            )
+
+        await _add_worker_shutdown_message(
+            (await self._request_result_queues)[0], num_workers
+        )
+
+    async def get_worker_function(
+        self, user_function: Callable[[_T], _U]
+    ) -> Callable[[str, str], None]:
+        if self._request_result_queues is None:
+            raise ValueError(
+                "Must call setup_and_add_tasks before calling get_worker_function"
+            )
+
+        request_queue, result_queue = await self._request_result_queues
+
+        return functools.partial(
+            worker_function,
+            user_function,
+            request_queue,
+            result_queue,
+        )
+
+    async def receive_task_results(
+        self, *, stop_receiving: asyncio.Event, workers_done: asyncio.Event
+    ) -> AsyncIterable[Tuple[int, int, ProcessState]]:
+        if self._request_result_queues is None:
+            raise ValueError(
+                "Must call setup_and_add_tasks before calling receive_task_results"
+            )
+
+        return get_results_unordered(
+            (await self._request_result_queues)[1],
+            self._location,
+            stop_receiving,
+            workers_done,
+        )
+
+    async def retry_task(self, task_id: int, attempts_so_far: int) -> None:
+        if self._tasks is None or self._request_result_queues is None:
+            raise ValueError("Must call setup_and_add_tasks before calling retry_task")
+
+        await _retry_task(
+            (await self._request_result_queues)[0],
+            task_id,
+            attempts_so_far + 1,
+            self._tasks[task_id],
+        )
