@@ -4,31 +4,32 @@ from __future__ import annotations
 import asyncio
 import dataclasses
 import datetime
-import functools
 import os
 import pickle
 import time
 import traceback
-import uuid
 from typing import (
-    TYPE_CHECKING,
     Any,
     AsyncIterable,
     Callable,
     Iterable,
+    List,
     Optional,
-    Sequence,
     Tuple,
     TypeVar,
+    cast,
 )
 
-from meadowrun.azure_integration.azure_instance_allocation import AzureInstanceRegistrar
 from meadowrun.azure_integration.azure_meadowrun_core import (
     ensure_meadowrun_storage_account,
-    get_default_location,
     record_last_used,
 )
-from meadowrun.azure_integration.azure_ssh_keys import ensure_meadowrun_key_pair
+from meadowrun.azure_integration.mgmt_functions.azure_constants import (
+    GRID_TASK_QUEUE,
+    QUEUE_NAME_TIMESTAMP_FORMAT,
+    _REQUEST_QUEUE_NAME_PREFIX,
+    _RESULT_QUEUE_NAME_PREFIX,
+)
 from meadowrun.azure_integration.mgmt_functions.azure_core.azure_rest_api import (
     azure_rest_api,
 )
@@ -38,18 +39,7 @@ from meadowrun.azure_integration.mgmt_functions.azure_core.azure_storage_api imp
     queue_receive_messages,
     queue_send_message,
 )
-from meadowrun.azure_integration.mgmt_functions.azure_constants import (
-    GRID_TASK_QUEUE,
-    QUEUE_NAME_TIMESTAMP_FORMAT,
-    _REQUEST_QUEUE_NAME_PREFIX,
-    _RESULT_QUEUE_NAME_PREFIX,
-)
-from meadowrun.instance_allocation import allocate_jobs_to_instances
-
-if TYPE_CHECKING:
-    from meadowrun.instance_selection import ResourcesInternal
 from meadowrun.meadowrun_pb2 import GridTask, GridTaskStateResponse, ProcessState
-from meadowrun.run_job_core import RunMapHelper
 from meadowrun.shared import pickle_exception
 
 _T = TypeVar("_T")
@@ -65,16 +55,6 @@ class Queue:
     storage_account: StorageAccount
 
 
-async def create_queues_and_add_tasks(
-    location: str, tasks: Iterable[Any]
-) -> Tuple[Queue, Queue]:
-    job_id = str(uuid.uuid4())
-    print(f"The current run_map's id is {job_id}")
-    request_queue, result_queue = await _create_queues_for_job(job_id, location)
-    await _add_tasks(request_queue, tasks)
-    return request_queue, result_queue
-
-
 async def _create_queues_for_job(job_id: str, location: str) -> Tuple[Queue, Queue]:
     storage_account = await ensure_meadowrun_storage_account(location, "create")
     now = datetime.datetime.utcnow().strftime(QUEUE_NAME_TIMESTAMP_FORMAT)
@@ -99,6 +79,9 @@ async def _create_queues_for_job(job_id: str, location: str) -> Tuple[Queue, Que
     )
 
 
+_WORKER_SHUTDOWN_MESSAGE = b"worker-shutdown"
+
+
 async def _add_tasks(request_queue: Queue, tasks: Iterable[Any]) -> None:
     await asyncio.wait(
         [
@@ -106,7 +89,7 @@ async def _add_tasks(request_queue: Queue, tasks: Iterable[Any]) -> None:
                 request_queue.storage_account,
                 request_queue.queue_name,
                 GridTask(
-                    task_id=i, pickled_function_arguments=pickle.dumps(task)
+                    task_id=i, attempt=1, pickled_function_arguments=pickle.dumps(task)
                 ).SerializeToString(),
             )
             for i, task in enumerate(tasks)
@@ -114,22 +97,54 @@ async def _add_tasks(request_queue: Queue, tasks: Iterable[Any]) -> None:
     )
 
 
-async def _get_task(request_queue: Queue, result_queue: Queue) -> Optional[GridTask]:
-    messages = await queue_receive_messages(
+async def _retry_task(
+    request_queue: Queue, task_id: int, attempt: int, task: Any
+) -> None:
+    await queue_send_message(
         request_queue.storage_account,
         request_queue.queue_name,
-        visibility_timeout_secs=5,
+        GridTask(
+            task_id=task_id,
+            attempt=attempt,
+            pickled_function_arguments=pickle.dumps(task),
+        ).SerializeToString(),
     )
 
-    # there was nothing in the queue
-    if len(messages) == 0:
-        return None
+
+async def _add_worker_shutdown_message(request_queue: Queue, num_workers: int) -> None:
+    # TODO I think these could be sent in a single API call
+    await asyncio.wait(
+        [
+            queue_send_message(
+                request_queue.storage_account,
+                request_queue.queue_name,
+                _WORKER_SHUTDOWN_MESSAGE,
+            )
+            for _ in range(num_workers)
+        ]
+    )
+
+
+async def _get_task(request_queue: Queue, result_queue: Queue) -> Optional[GridTask]:
+    while True:
+        messages = await queue_receive_messages(
+            request_queue.storage_account,
+            request_queue.queue_name,
+            visibility_timeout_secs=5,
+        )
+
+        if len(messages) > 0:
+            break
 
     if len(messages) > 1:
         raise ValueError(
             "queue_receive_messages returned more than 1 message unexpectedly"
         )
 
+    # if we got a worker done message, return None
+    if messages[0].message_content == _WORKER_SHUTDOWN_MESSAGE:
+        return None
+
     task = GridTask()
     task.ParseFromString(messages[0].message_content)
 
@@ -159,32 +174,36 @@ async def _complete_task(
     result_queue: Queue,
     task: GridTask,
     process_state: ProcessState,
-    public_address: str,
-    worker_id: int,
 ) -> None:
     await queue_send_message(
         result_queue.storage_account,
         result_queue.queue_name,
         GridTaskStateResponse(
-            task_id=task.task_id, process_state=process_state
+            task_id=task.task_id, attempt=task.attempt, process_state=process_state
         ).SerializeToString(),
     )
 
 
-async def worker_loop_async(
+async def worker_function_async(
     function: Callable[[Any], Any],
     request_queue: Queue,
     result_queue: Queue,
     public_address: str,
-    worker_id: int,
+    log_file_name: str,
 ) -> None:
     pid = os.getpid()
+    log_file_name = f"{public_address}:{log_file_name}"
 
     while True:
         task = await _get_task(request_queue, result_queue)
         if not task:
             break
 
+        print(
+            f"Meadowrun agent: About to execute task #{task.task_id}, attempt "
+            f"#{task.attempt}"
+        )
+
         try:
             result = function(pickle.loads(task.pickled_function_arguments))
         except Exception as e:
@@ -195,6 +214,7 @@ async def worker_loop_async(
                 pid=pid,
                 pickled_result=pickle_exception(e, pickle.HIGHEST_PROTOCOL),
                 return_code=0,
+                log_file_name=log_file_name,
             )
         else:
             process_state = ProcessState(
@@ -202,33 +222,36 @@ async def worker_loop_async(
                 pid=pid,
                 pickled_result=pickle.dumps(result, protocol=pickle.HIGHEST_PROTOCOL),
                 return_code=0,
+                log_file_name=log_file_name,
             )
 
-        await _complete_task(
-            result_queue, task, process_state, public_address, worker_id
+        print(
+            f"Meadowrun agent: Completed task #{task.task_id}, attempt {task.attempt}, "
+            f"state {ProcessState.ProcessStateEnum.Name(process_state.state)}"
         )
+        await _complete_task(result_queue, task, process_state)
 
 
-def worker_loop(
+def worker_function(
     function: Callable[[Any], Any],
     request_queue: Queue,
     result_queue: Queue,
     public_address: str,
-    worker_id: int,
+    log_file_name: str,
 ) -> None:
     asyncio.run(
-        worker_loop_async(
-            function, request_queue, result_queue, public_address, worker_id
+        worker_function_async(
+            function, request_queue, result_queue, public_address, log_file_name
         )
     )
 
 
 async def get_results_unordered(
     result_queue: Queue,
-    num_tasks: int,
     location: str,
-    workers_done: Optional[asyncio.Event] = None,
-) -> AsyncIterable[Tuple[int, ProcessState]]:
+    stop_receiving: asyncio.Event,
+    all_workers_exited: asyncio.Event,
+) -> AsyncIterable[Tuple[int, int, ProcessState]]:
 
     # TODO currently, we get back messages saying that a task is running on a particular
     # worker. We don't really do anything with these messages, but eventually we should
@@ -237,89 +260,59 @@ async def get_results_unordered(
     num_tasks_running, num_tasks_completed = 0, 0
     t0 = None
     updated = True
-    while num_tasks_completed < num_tasks:
+    stop_receiving_wait_task = asyncio.create_task(stop_receiving.wait())
+    all_workers_exited_task = asyncio.create_task(all_workers_exited.wait())
+    while not stop_receiving.is_set():
         if updated or t0 is None or time.time() - t0 > 20:
-            # log this message every 20 seconds, or whenever there's an update
             t0 = time.time()
             updated = False
-            print(
-                f"Waiting for grid tasks. Requested: {num_tasks}, "
-                f"running: {num_tasks_running}, "
-                f"completed: {num_tasks_completed}"
-            )
             await record_last_used(GRID_TASK_QUEUE, result_queue.queue_job_id, location)
 
-        for message in await queue_receive_messages(
-            result_queue.storage_account,
-            result_queue.queue_name,
-            visibility_timeout_secs=10,
-            num_messages=32,  # the max number of messages to get at once
-        ):
-            updated = True
-
-            task_result = GridTaskStateResponse()
-            task_result.ParseFromString(message.message_content)
-
-            if (
-                task_result.process_state.state
-                == ProcessState.ProcessStateEnum.RUN_REQUESTED
-            ):
-                num_tasks_running += 1
-            else:
-                num_tasks_running -= 1
-                num_tasks_completed += 1
-                yield task_result.task_id, task_result.process_state
-
-            await queue_delete_message(
+        receive_messages_task = asyncio.create_task(
+            queue_receive_messages(
                 result_queue.storage_account,
                 result_queue.queue_name,
-                message.message_id,
-                message.pop_receipt,
+                visibility_timeout_secs=10,
+                num_messages=32,  # the max number of messages to get at once
             )
-
-
-async def prepare_azure_vm_run_map(
-    function: Callable[[_T], _U],
-    tasks: Sequence[_T],
-    location: Optional[str],
-    resources_required_per_task: ResourcesInternal,
-    num_concurrent_tasks: int,
-    ports: Optional[Sequence[str]],
-) -> RunMapHelper:
-    """This code is tightly coupled with run_map"""
-    if not location:
-        location = get_default_location()
-
-    key_pair_future = asyncio.create_task(ensure_meadowrun_key_pair(location))
-    queues_future = asyncio.create_task(create_queues_and_add_tasks(location, tasks))
-
-    async with AzureInstanceRegistrar(location, "create") as instance_registrar:
-        if ports:
-            raise NotImplementedError(
-                "Support for opening ports on Azure is not yet implemented, please "
-                "create an issue at https://github.com/meadowdata/meadowrun/issues"
-            )
-        allocated_hosts = await allocate_jobs_to_instances(
-            instance_registrar,
-            resources_required_per_task,
-            num_concurrent_tasks,
-            location,
-            ports,
         )
+        await asyncio.wait(
+            cast(
+                List[asyncio.Task],
+                [
+                    receive_messages_task,
+                    stop_receiving_wait_task,
+                    all_workers_exited_task,
+                ],
+            ),
+            return_when=asyncio.FIRST_COMPLETED,
+        )
+        if stop_receiving_wait_task.done() or all_workers_exited_task.done():
+            return
+        else:
+            for message in receive_messages_task.result():
+                updated = True
 
-    private_key, public_key = await key_pair_future
-    request_queue, result_queue = await queues_future
+                task_result = GridTaskStateResponse()
+                task_result.ParseFromString(message.message_content)
 
-    return RunMapHelper(
-        location,
-        allocated_hosts,
-        worker_function=functools.partial(
-            worker_loop, function, request_queue, result_queue
-        ),
-        ssh_username="meadowrunuser",
-        ssh_private_key=private_key,
-        num_tasks=len(tasks),
-        process_state_futures=functools.partial(
-            get_results_unordered, result_queue, len(tasks), location
-        ),
-    )
+                if (
+                    task_result.process_state.state
+                    == ProcessState.ProcessStateEnum.RUN_REQUESTED
+                ):
+                    num_tasks_running += 1
+                else:
+                    num_tasks_running -= 1
+                    num_tasks_completed += 1
+                    yield (
+                        task_result.task_id,
+                        task_result.attempt,
+                        task_result.process_state,
+                    )
+
+                await queue_delete_message(
+                    result_queue.storage_account,
+                    result_queue.queue_name,
+                    message.message_id,
+                    message.pop_receipt,
+                )
