@@ -8,6 +8,7 @@ import uuid
 from typing import (
     TYPE_CHECKING,
     Any,
+    AsyncIterable,
     Dict,
     Generic,
     Iterable,
@@ -19,7 +20,6 @@ from typing import (
     TypeVar,
 )
 
-
 from meadowrun.instance_selection import (
     CloudInstance,
     ResourcesInternal,
@@ -29,6 +29,7 @@ from meadowrun.shared import assert_is_not_none
 
 if TYPE_CHECKING:
     from types import TracebackType
+    from meadowrun.run_job_core import AllocVM
 
 
 @dataclasses.dataclass
@@ -185,7 +186,8 @@ class InstanceRegistrar(abc.ABC, Generic[_TInstanceState]):
         self,
         resources_required_per_task: ResourcesInternal,
         num_concurrent_tasks: int,
-        region_name: str,
+        alloc_cloud_instances: AllocVM,
+        abort: Optional[asyncio.Event],
     ) -> Sequence[CloudInstance]:
         """
         This isn't per se part of the "instance registration" process, but it's helpful
@@ -367,10 +369,11 @@ async def _choose_existing_instances(
 
 async def _launch_new_instances(
     instance_registrar: InstanceRegistrar,
-    resources_required_per_task: ResourcesInternal,
+    instance_type_resources_required_per_task: ResourcesInternal,
     num_concurrent_tasks: int,
-    region_name: str,
+    alloc_cloud_instance: AllocVM,
     original_num_concurrent_tasks: int,
+    abort: Optional[asyncio.Event],
 ) -> Tuple[Dict[str, List[str]], Dict[str, CloudInstance]]:
     """
     Chooses the cheapest instances to launch that can run the specified jobs, launches
@@ -379,11 +382,20 @@ async def _launch_new_instances(
     Returns two dictionaries, {public_address: [job_ids]}, {public_address:
     cloud_instance}. The keys of the two dictionaries will be the same.
 
+    The CloudInstance objects returned will have a resources field that includes both
+    the instance type resources (like CPU and memory) that were used to select an
+    instance type as well as the "runtime resources" provided by alloc_cloud_instance
+    (like the AMI id and subnet ID) which describe how the instance was actaully
+    launched.
+
     original_num_jobs is only needed to produce more coherent logging.
     """
 
     instances = await instance_registrar.launch_instances(
-        resources_required_per_task, num_concurrent_tasks, region_name
+        instance_type_resources_required_per_task,
+        num_concurrent_tasks,
+        alloc_cloud_instance,
+        abort,
     )
 
     description_strings = []
@@ -395,6 +407,9 @@ async def _launch_new_instances(
     for instance in instances:
         # just to make the code more readable
         instance_info = instance.instance_type.instance_type
+        all_resources = instance_info.resources.combine(
+            alloc_cloud_instance.get_runtime_resources()
+        )
 
         # the number of jobs to allocate to this instance
         num_allocated_jobs = min(
@@ -408,11 +423,13 @@ async def _launch_new_instances(
             instance.public_dns_name,
             instance.name,
             assert_is_not_none(
-                instance_info.resources.subtract(
-                    resources_required_per_task.multiply(num_allocated_jobs)
+                all_resources.subtract(
+                    instance_type_resources_required_per_task.multiply(
+                        num_allocated_jobs
+                    )
                 )
             ),
-            [(job_id, resources_required_per_task) for job_id in job_ids],
+            [(job_id, instance_type_resources_required_per_task) for job_id in job_ids],
         )
 
         allocated_jobs[instance.public_dns_name] = job_ids
@@ -444,9 +461,10 @@ async def allocate_jobs_to_instances(
     instance_registrar: InstanceRegistrar,
     resources_required_per_task: ResourcesInternal,
     num_concurrent_tasks: int,
-    region_name: str,
+    alloc_cloud_instance: AllocVM,
     ports: Optional[Sequence[str]],
-) -> Dict[str, List[str]]:
+    abort: Optional[asyncio.Event],
+) -> AsyncIterable[Dict[str, List[str]]]:
     """
     This function first tries to re-use existing instances, and if necessary launches
     the cheapest possible new instances that have the requested resources.
@@ -458,30 +476,83 @@ async def allocate_jobs_to_instances(
         instance_registrar.authorize_current_ip()
     )
 
-    allocated_jobs, allocated_existing_instances = await _choose_existing_instances(
+    (
+        existing_instances_allocated_jobs,
+        allocated_existing_instances,
+    ) = await _choose_existing_instances(
         instance_registrar,
-        resources_required_per_task,
+        resources_required_per_task.combine(
+            alloc_cloud_instance.get_runtime_resources()
+        ),
         num_concurrent_tasks,
     )
 
+    if len(existing_instances_allocated_jobs) > 0:
+        await instance_registrar.open_ports(
+            ports, allocated_existing_instances.values(), ()
+        )
+        await authorize_current_ip_task  # this should almost always be complete by now
+        yield existing_instances_allocated_jobs
+
     num_concurrent_tasks_remaining = num_concurrent_tasks - sum(
-        len(jobs) for jobs in allocated_jobs.values()
+        len(jobs) for jobs in existing_instances_allocated_jobs.values()
     )
 
-    allocated_new_instances: Dict[str, CloudInstance] = {}
     if num_concurrent_tasks_remaining > 0:
-        allocated_new_jobs, allocated_new_instances = await _launch_new_instances(
+        (
+            new_instances_allocated_jobs,
+            allocated_new_instances,
+        ) = await _launch_new_instances(
             instance_registrar,
             resources_required_per_task,
             num_concurrent_tasks_remaining,
-            region_name,
+            alloc_cloud_instance,
             num_concurrent_tasks,
+            abort,
         )
-        allocated_jobs.update(allocated_new_jobs)
+        # TODO if this fails and we have existing instances, we should just carry on
+        # with those
 
-    await instance_registrar.open_ports(
-        ports, allocated_existing_instances.values(), allocated_new_instances.values()
-    )
-    await authorize_current_ip_task
+        await instance_registrar.open_ports(ports, (), allocated_new_instances.values())
+        if len(existing_instances_allocated_jobs) == 0:
+            # we need to await this if and only if we didn't await it before
+            await authorize_current_ip_task
 
-    return allocated_jobs
+        yield new_instances_allocated_jobs
+
+
+async def allocate_single_job_to_instance(
+    instance_registrar: InstanceRegistrar,
+    resources_required: ResourcesInternal,
+    alloc_cloud_instance: AllocVM,
+    ports: Optional[Sequence[str]],
+) -> Tuple[str, str]:
+    """
+    A thin wrapper around allocate_jobs_to_instances with a simpler return type for when
+    you only need one job.
+    """
+
+    result = None
+
+    async for allocated_hosts in allocate_jobs_to_instances(
+        instance_registrar, resources_required, 1, alloc_cloud_instance, ports, None
+    ):
+        if len(allocated_hosts) == 0:
+            pass
+        elif len(allocated_hosts) > 1:
+            raise ValueError(f"Requested 1 host, but got back {len(allocated_hosts)}")
+        else:
+            if result is None:
+                host, job_ids = list(allocated_hosts.items())[0]
+                if len(job_ids) != 1:
+                    raise ValueError(
+                        f"Requested 1 job allocation but got back {len(job_ids)}"
+                    )
+                result = host, job_ids[0]
+            else:
+                raise ValueError("Requested 1 host, but got back more than one")
+
+    if result is None:
+        raise ValueError("Requested 1 host, but did not get any back")
+
+    return result
